/**
   ##############################################################################
   ##                                                                          ##
   ## EnvInfo                                                                  ##
   ##                                                                          ##
   ## Copyright (C) 2009  Frederic Roudaut  <frederic.roudaut@free.fr>         ##
   ##                                                                          ##
   ## This class is an adaption of the one done by Nicolas Richasse for his    ##
   ## Java Iperf frontend.                                                     ##
   ##                                                                          ##
   ##                                                                          ##
   ## This program is free software: you can redistribute it and/or modify     ##
   ## it under the terms of the GNU General Public License as published by     ##
   ## the Free Software Foundation, either version 3 of the License, or        ##
   ## (at your option) any later version.                                      ##
   ##                                                                          ##
   ## This program is distributed in the hope that it will be useful,          ##
   ## but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
   ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            ##
   ## GNU General Public License for more details.                             ##
   ##                                                                          ##
   ## You should have received a copy of the GNU General Public License        ##
   ## along with this program.  If not, see <http://www.gnu.org/licenses/>.    ##
   ##                                                                          ##
   ##                                                                          ##
   ##############################################################################
**/

package com.envInfo;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridLayout;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;

import javax.swing.JLabel;
import javax.swing.JPanel;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.AxisLocation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.CombinedDomainXYPlot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;


/**
 * Class used to draw the JVM Memory status charts.
 * <br/><br/>
 * This class is an adaption of the one done by Nicolas Richasse for his   
 * Java Iperf frontend.
 *
 **/
public class MemoryChartPanel extends AbstractChartPanel
    implements Runnable
{

	 private static final long serialVersionUID = 1L;
	 
    /**
     * Class representing a single series of Memory data.
     * 
     **/
    private class SeriesData
    {
	public String	seriesId;
	public String	memoryLegend;
	public JLabel	seriesLabel;
	public XYSeries	memorySeries;
	public Color	seriesColor;
	public String	printfMemoryValueExpression;

	/**
	 * Constructor.
	 * @param seriesId ID.
	 * @param memoryLegend Memory Legend.
	 * @param seriesColor Color for the chart.
	 * @param printfMemoryValueExpression Content to print under the Chart.
	 *
	 **/
	public SeriesData (String seriesId, String memoryLegend, Color seriesColor, 
			   String printfMemoryValueExpression)
	{
	    this.seriesId = seriesId;
	    this.memoryLegend = memoryLegend;
	    this.memorySeries = new XYSeries(memoryLegend);
	    this.seriesLabel = new JLabel(memoryLegend);
	    this.seriesColor = seriesColor;
	    this.printfMemoryValueExpression = printfMemoryValueExpression;
	    seriesLabel.setForeground(seriesColor);
	}
    }

    private int	proportion;
    private CombinedDomainXYPlot graphSet;
    private XYItemRenderer memoryRenderer;
    private XYSeriesCollection memoryCollection;
    private JPanel panelTextStats = new JPanel(new GridLayout(0, 4));
    private JLabel labelDate = new JLabel(" ");
    private double delayInSeconds;
    private SimpleDateFormat sdf = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss");
    private HashMap<String, SeriesData>	seriesData = new HashMap<String, SeriesData>();
    private Thread timeThread = null;
    private Color backgroundColor, foregroundColor, gridColor;
    private String memoryUnit;
    private double timeWindow;
	
    /**
     * Constructor for a set of Memory Charts (ie a set of memory series).
     * @param title Title of the Memory Charts set. 
     * @param memoryUnit Memory Unit.
     * @param timeAxisLabel Time Axis Label (X axis).
     * @param memoryValueAxisLabel Memory Value Axis Label (Y Axis).
     * @param delayInSeconds Delay between the date printing above the charts set.
     * @param timeWindow Time Windows for the charts.
     * @param backgroundColor BackgroundColor.
     * @param foregroundColor ForegroundColor.
     * @param gridColor GridColor.
     *
     **/
    public MemoryChartPanel(String title, String memoryUnit, String timeAxisLabel, 
			    String memoryValueAxisLabel, double delayInSeconds, double timeWindow, 
			    Color backgroundColor, Color foregroundColor, Color gridColor)
    {
	this.delayInSeconds = delayInSeconds;
	this.proportion = 1;
	this.backgroundColor = backgroundColor;
	this.foregroundColor = foregroundColor;
	this.gridColor = gridColor;

	reconfigure(title, memoryUnit, timeAxisLabel, memoryValueAxisLabel, timeWindow);
    }

    /**
     *
     * Start a Time Thread.
     *
     */    
    public void start()
    {
	if (timeThread == null)
	    {
		this.timeThread = new Thread(this);
		timeThread.start();
	    }
    }

    /**
     *
     * Check if a Memory Data Series Exists.
     * @param seriesId A serie ID. 
     * @return true if the serie exists, false otherwise.
     *
     */
    private boolean seriesExists(String seriesId)
    {
	if (seriesData.keySet().contains(seriesId))
	    {
		return true;
	    }
	else
	    {
		return false;
	    }
    }

    /**
     *
     * Add a Memory Data Series.
     * @param seriesId A serie ID. 
     * @param seriesLegend the Serie Legend
     * @param seriesColor a color for the associated chart.
     *
     */
    public void maybeAddNewSeries(String seriesId, String seriesLegend, Color seriesColor)
    {
	maybeAddNewSeries(seriesId, seriesLegend, seriesColor, "%4.2f");
    }

    /**
     *
     * Add a Memory Data Series.
     * @param seriesId A serie ID. 
     * @param seriesLegend the Serie Legend
     * @param seriesColor a color for the associated chart.
     * @param printfMemoryValueExpression Content to print under the Charts.
     *
     */
    public synchronized void maybeAddNewSeries(String seriesId, String seriesLegend, 
					       Color seriesColor, String printfMemoryValueExpression)
    {
	if (!seriesExists(seriesId))
	    {
		SeriesData data = new SeriesData(seriesId, seriesLegend, seriesColor, printfMemoryValueExpression);
		seriesData.put(seriesId, data);
			
		memoryCollection.addSeries(data.memorySeries);
		memoryRenderer.setSeriesPaint(memoryCollection.getSeriesCount()-1, data.seriesColor);

		panelTextStats.add(data.seriesLabel);
	    }
    }

    /**
     *
     * Reconfigure a Memory Data Charts set.
     * @param title Title of the Memory Charts set. 
     * @param memoryUnit Memory Unit.
     * @param timeAxisLabel Time Axis Label.
     * @param memoryValueAxisLabel Memory Value Axis Label.
     * @param timeWindow Time windows for the charts.
     *
     */
    public void reconfigure(String title, String memoryUnit, String timeAxisLabel, String memoryValueAxisLabel, 
			    double timeWindow)
    {
	this.memoryUnit = memoryUnit;
	this.timeWindow = timeWindow;
		
	// reset the content pane
	this.removeAll();
	panelTextStats.removeAll();
	SeriesColorGenerator.reset();

	seriesData.clear();

	// creation of the chart group
	graphSet = new CombinedDomainXYPlot(new NumberAxis(timeAxisLabel));
	// space between charts
	graphSet.setGap(10.0);
	// creation of the jFreeChart
	jFreeChart = new JFreeChart(null, JFreeChart.DEFAULT_TITLE_FONT, graphSet, false);
	jFreeChart.setBackgroundPaint(backgroundColor);
	if (title != null)
	    {
		jFreeChart.setTitle(title);
		jFreeChart.getTitle().setPaint(foregroundColor);
	    }
	// creation of the chart panel
	chartPanel = new ChartPanel(jFreeChart);
	chartPanel.setBackground(backgroundColor);
	// creation of the series set
	memoryCollection = new XYSeriesCollection();
	// creation of the renderer
	memoryRenderer = new XYLineAndShapeRenderer();
	// set the UI presentation
	NumberAxis rangeAxis = new NumberAxis(memoryValueAxisLabel);
	rangeAxis.setLabelPaint(foregroundColor);

	// creation of the memory plot
	XYPlot memoryPlot = new XYPlot(memoryCollection, null, rangeAxis, memoryRenderer);
	memoryPlot.setRangeAxisLocation(AxisLocation.BOTTOM_OR_LEFT);
	memoryPlot.setDomainCrosshairVisible(false);
	memoryPlot.setRangeCrosshairVisible(false);
	memoryPlot.setBackgroundPaint(backgroundColor);
	memoryPlot.setDomainGridlinePaint(gridColor);
	memoryPlot.setRangeGridlinePaint(gridColor);

	// add the plot
	graphSet.add(memoryPlot, proportion);

	// set up the domain axis
	ValueAxis axis = memoryPlot.getDomainAxis();
	if (timeAxisLabel != null)
	    {
		axis.setTickLabelPaint(foregroundColor);
		axis.setLabel(timeAxisLabel);
		axis.setLabelPaint(foregroundColor);
	    }
	else
	    {
		axis.setVisible(false);
	    }
	axis.setAutoRange(true);
	axis.setFixedAutoRange((int)Math.min(timeWindow, 60));

	// set up the range axis
	axis = memoryPlot.getRangeAxis();
	axis.setTickLabelPaint(foregroundColor);

	// set up the hour&date label presentation
	labelDate.setHorizontalAlignment(JLabel.RIGHT);
	labelDate.setForeground(foregroundColor);

	panelTextStats.setBackground(backgroundColor);

	this.add(labelDate, BorderLayout.NORTH);
	this.add(chartPanel, BorderLayout.CENTER);
	this.add(panelTextStats, BorderLayout.SOUTH);

	this.setBackground(backgroundColor);
    }

    /**
     *
     * Add a new measurement to an existing memory data serie.
     * @param seriesId the sery ID.
     * @param measurement the measurement to add.
     *
     */    
    public void addSeriesMemoryMeasurement(String seriesId, MemoryMeasurement measurement)
    {
	SeriesData data = seriesData.get(seriesId);
	data.memorySeries.add(measurement.getEndTime(), measurement.getValue());
	data.seriesLabel.setText(String.format("<html><b>%s</b> " + data.printfMemoryValueExpression + "%s </html>", 
					       data.memoryLegend, measurement.getValue(),
					       " " + measurement.getUnits() +"&nbsp;&nbsp;"));
    }
    
    /**
     *
     * Start a Thread to update the date.
     *
     */    
    public void run()
    {
	while (true)
	    {
		try
		    {
			Date d = new Date();
			labelDate.setText(sdf.format(d));
			Thread.sleep((int) (delayInSeconds * 1000));
		    }
		catch (Exception ex)
		    {
		    }
	    }
    }
}
